__version__ = "0.1.0"

from .player_panel import PlayerPanel
from .player import Player
from .tttboard import TTTBoard,TTTBox, GameStatusNote

__all__ = ['TTTBoard', 'TTTBox', 'GameStatusNote', 'Player', 'PlayerPanel',"__version__"]


import argparse
from sys import version_info, exit
import random
from .__init__ import __version__

from textual.app import App
from textual.widgets import Footer, Header

from .tttboard import TTTBoard, GameStatusNote
from .player_panel import PlayerPanel
from .player import Player

def run(argv=None):
    parser = argparse.ArgumentParser(description="A tictactoe TUI.")
    parser.prog = "texttactoe"

    parser.add_argument("-v", "--version"   ,action="version", version=get_version(), help="display version information",)
    parser.add_argument("-p1", "--player1"  ,default="Player1", help="Name of Player1",)
    parser.add_argument("-p2", "--player2"  ,default="Player2", help="Name of Player2",)
    parser.add_argument("-c1", "--color1"   ,default="rgb(0,161,162)", help="Color of Player1",)
    parser.add_argument("-c2", "--color2"   ,default="yellow", help="Color of Player2",)
    parser.add_argument("-r", "--random"    ,action="store_true", help="Randomize player colors")

    args = parser.parse_args(argv)


    if args.random:
        vals =["red","green","blue","yellow","cyan","Magenta","bright_yellow","bright_cyan"]
        r1 = random.sample(vals,1)
        r2=r1
        while r1[0]==r2[0]:
            r2 = random.sample(vals,1)
        args.color1 =  r1[0]
        args.color2 =  r2[0]



    TextTacToe.run(
        title = "TextTacToe",
        log="textual.log",  
        log_verbosity=3,
        player1 = Player(args.player1,color=args.color1),
        player2 = Player(args.player2,color=args.color2),
        )


class TextTacToe(App):

    def __init__(self, *args, 
        player1, 
        player2, 
        **kwargs):

        self.player1 = player1
        self.player2 = player2
        self.players = [player1, player2]

        super().__init__(*args, **kwargs)

    async def on_load(self) -> None:
        """Bind keys with the app loads (but before entering application mode)"""
        await self.bind("q", "quit", "Quit")
        await self.bind("escape", "quit", "Quit")
        await self.bind("r", "reset_board", "Reset Game")

    async def on_mount(self) -> None:

        self.game_board = TTTBoard(self.player1,self.player2)
        self.footer = Footer()
        self.header = Header(style="white on black")
        self.info_panels = [PlayerPanel(name = "InfoPanel", player=p) for p in self.players]

        await self.view.dock(self.header,edge = "top")
        await self.view.dock(self.footer,edge="bottom")
        self.grid = await self.view.dock_grid(edge = "top")

        self.grid.add_column("col0", fraction=1, max_size=30)
        self.grid.add_column("col1", fraction=2)
        self.grid.add_row("row1", fraction=1, max_size = 7)
        self.grid.add_row("row2", fraction=1, max_size = 7)
        self.grid.add_row("row3", fraction=1, max_size=7)

        self.grid.add_areas(
                panel1  = "col0,row1",
                panel2  = "col0,row2",
                board   = "col1,row1-start|row3-end",
        )
        self.grid.place(panel1  = self.info_panels[0])
        self.grid.place(panel2  = self.info_panels[1])
        self.grid.place(board   = self.game_board)

    async def action_reset_board(self):
        self.game_board.reset_game()

    def update_panels(self, player):
        """Update player panel info"""
        for p in self.info_panels:
            if p.player.name == player.name:
                p.new_content(player)

    async def handle_game_status_note(self, message: GameStatusNote):
        # handle win
        if message.winner:
            self.update_panels(message.winner)
        else:
            pass
            # TODO: Handle Ties

def get_version():
    python_version = f"{version_info.major}.{version_info.minor}.{version_info.micro}"
    return "\n".join([f"texttactoe {__version__} [Python {python_version}]",])

if __name__ == "__main__":
    run()


from itertools import cycle

from textual.widgets import Button
from textual.widget import Widget
from textual.reactive import Reactive
from textual.views import GridView
from textual.message import Message

from rich.panel import Panel

class GameStatusNote(Message):
    """A Message meant for the App to update game info"""
    def __init__(self, sender, winner=None, ) -> None:
        super().__init__(sender)
        self.winner = winner

class TTTBox(Widget):
    """A interactable box in TicTacToe widget"""

    # winner: denotes whether this box is a part of a winning 3-in-a-row
    # disable: denotes whether this box is disbaled for interaction, happens on game_over
    mouse_over = Reactive(False)
    is_selected = Reactive(False)
    winner = Reactive(False)
    color = ""
    disable = False

    def __init__(self, board, name=None):
        super().__init__(name=name)
        self.board = board

    def render(self) -> Panel:
        if not self.is_selected:
            if self.mouse_over:
                self.color = self.board.current_turn.color
            else:
                self.color = "grey35"
        else:
            if self.winner:
                self.color = "bright_white"

        r = Button("", style=f"on {self.color}")
        return r

    def toggle_disable(self):
        """Used to disable tiles on game over"""
        self.disable = not self.disable
        self.log(f"toggling myslef {self}")

    def reset_tttbox(self) -> None:
        """Resets tile to prepare for next game"""
        self.toggle_disable()
        self.is_selected = False
        self.winner = False
        self.mouse_over = False

    def on_enter(self) -> None:
        if not self.disable:
            self.mouse_over = True

    def on_leave(self) -> None:
        if not self.disable:
            self.mouse_over = False

    def on_click(self) -> None:
        if not self.disable:
            if not self.is_selected:
                self.is_selected=True
                self.board.react_box_click()

class Player():

    def __init__(self, name:str, color:str=None):
        self.name = name
        self.color = color
        self.wins = 0
        self.ties = 0
        self.losses = 0
        self.streak = 0

    def add_win(self):
        self.wins+=1

class TTTBoard(GridView):
    """TicTacToe Board widget"""

    #set of possoble states after a click is received 
    CONTINUE = 0
    WIN = 1
    TIE = 2

    rows = 3
    columns = 3
    show_end_panel = Reactive(False)

    def __init__(self, player1: Player = None, player2:Player = None):
        super().__init__()

        default_pl = Player("Player1",color="red")
        default_p2 = Player("Player2",color="blue")
        if not player1:
            self.player1 = default_pl
            self.player2 = default_p2
        elif (not player2):
            self.player1 = player1
            self.player2 = default_p2
        else:
            self.player1 = player1
            self.player2 = player2
        
        self.current_turn = self.player1
        self.player_list = [self.player1, self.player2]
        self.player_cycle = cycle(self.player_list)

    # TODO: test speed if made async 
    def on_mount(self) -> None:
        self.init_game()
        self.grid.set_align("center", "top")    
        self.grid.set_gap(2,1)
        self.grid.set_gutter(column=3,row=1)

        # add rows and columns with repeat arg
        self.grid.add_column(name="col", min_size=2, max_size=10,repeat=3)
        self.grid.add_row(name="row", min_size=2, max_size=5,repeat=3)
        
        self.board = [TTTBox(self) for _ in range(self.rows*self.columns)]

        self.grid.place(*self.board)

    def display_win(self, indexes):
        """Show winning 3-in-a-row"""
        self.log(f"WIN INDEXES: {indexes}")
        for r,c in indexes:
            tile = self.board_access(r, c, self.rows)
            tile.winner = True


    def win_indexes(self, n):
        """Compute Win indexes for TicTacToe"""
        # Rows
        for r in range(n):
            yield [(r, c) for c in range(n)]
        # Columns
        for c in range(n):
            yield [(r, c) for r in range(n)]
        # Diagonal top left to bottom right
        yield [(i, i) for i in range(n)]
        # Diagonal top right to bottom left
        yield [(i, n - 1 - i) for i in range(n)]

    def board_access(self,r,c,n):
        """"Helper function, calculates position in 1D array of board tiles"""
        return self.board[((r*n)+c)]

    def is_winner(self):
        """Checks if current board arengement is a winning combination"""
        for indexes in self.win_indexes(self.rows):
            if all((self.board_access(r,c,self.rows).color == self.current_turn.color )for r, c in indexes):
                return self.WIN, indexes

        selected = 0
        for tile in self.board:
            if tile.is_selected:
                selected+=1
        if selected ==self.rows*self.columns:
            return self.TIE,[]

        return self.CONTINUE,[]

    def init_game(self):
        """Initializes Game turns"""
        self.switch_turns()

    def switch_turns(self):
        """Switched turns to next player in the cycle"""
        self.current_turn = next(self.player_cycle)

    def reset_game(self):
        """Resets the game board """
        for box in self.board:
            box.reset_tttbox() 
        self.won = False
        self.switch_turns()

    def toggle_disable(self):
        """Toggle board interactability"""
        for tile in self.board:
            tile.toggle_disable()

    def react_box_click(self):
        """Handle a TTTBoxClick"""
        # TODO: this could possbly be replaced by watchers, action, and compute functions
        status,indexes = self.is_winner()
        if status is self.WIN:
            self.current_turn.add_win()
            self.emit_no_wait(GameStatusNote(self,self.current_turn))
            self.display_win(indexes)
            self.toggle_disable()
            # TODO: Show win Panel
        elif status is self.TIE:
            self.emit_no_wait(GameStatusNote(self))
            self.toggle_disable()
            # self.display_tie()
            # TODO: Show win Panel
        else:
            self.switch_turns()



from textual import events
from textual.widget import Widget
from textual.reactive import Reactive

from rich.panel import Panel
from rich.align import Align
from rich import box
from rich.console import RenderableType

class PlayerPanel(Widget):

    mouse_over: Reactive[bool] = Reactive(False)
    style: Reactive[str] = Reactive("")
    content = Reactive("")
    player = None

    def __init__(self, *, name= None,player =None) -> None:
        super().__init__(name=name)
        self.player = player
        self.content =  f"Wins: {self.player.wins}"

    def render(self) -> RenderableType:
        return Panel(
                Align.left(self.content),
                title=f"{self.player.name}",
                border_style="bright_white" if self.mouse_over else f"{self.player.color}",
                box=box.ROUNDED,
                )

    def new_content(self, player):
        """Update player panel content"""
        self.player = player
        self.content = f"Wins: {self.player.wins}"

    async def on_focus(self, event: events.Focus) -> None:
        self.has_focus = True

    async def on_blur(self, event: events.Blur) -> None:
        self.has_focus = False

    async def on_enter(self, event: events.Enter) -> None:
        self.mouse_over = True

    async def on_leave(self, event: events.Leave) -> None:
        self.mouse_over = False




class Player():

    def __init__(self, name:str, color:str=None):
        self.name = name
        self.color = color
        self.wins = 0
        self.ties = 0
        self.losses = 0
        self.streak = 0

    def add_win(self):
        self.wins+=1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['texttactoe']

package_data = \
{'': ['*']}

install_requires = \
['textual>=0.1.18,<0.2.0']

entry_points = \
{'console_scripts': ['texttactoe = texttactoe.__main__:run']}

setup_kwargs = {
    'name': 'texttactoe',
    'version': '0.1.0',
    'description': 'A TUI to play tictactoe.',
    'long_description': '# TextTacToe\nThis is an TUI App for playing TicTacToe built using [Textual](https://github.com/willmcgugan/textual/) in Python3.\n\n![View Of TextTacToe](img/TextTacToe.gif)\n\n>**NOTE:** Due to the fact that textual t is still in the develepment stage, I reccomend you install this only in virtual env. This app will probably change heavily as textual changes.\n\n## Dependencies\n\nTextual:\n```\npython -m pip install textual \n```\n\n## Install\n\n```\npython -m pip install texttactoe\n```\n## Run\n\nIf installed correctly, you should be able to run 1 of 2 ways, as a python module or a standalone script:\n\n```\npython -m texttactoe\n```\nor\n```\ntexttactoe\n```\n The following options are available for customization. Note that the color options currently do not check whether they are a valid colors. We use textual/rich to customize the color and so it mus be a valid choice from these frameworks.\n```\nusage: texttactoe [-h] [-v] [-p1 PLAYER1] [-p2 PLAYER2] [-c1 COLOR1] [-c2 COLOR2] [-r]\n\nA tictactoe TUI.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -v, --version         display version information\n  -p1 PLAYER1, --player1 PLAYER1\n                        Name of Player1\n  -p2 PLAYER2, --player2 PLAYER2\n                        Name of Player2\n  -c1 COLOR1, --color1 COLOR1\n                        Color of Player1\n  -c2 COLOR2, --color2 COLOR2\n                        Color of Player2\n  -r, --random          Randomize player colors\n```',
    'author': 'Jose Medina',
    'author_email': 'jose.amedina96@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/JoseAndresMedina/TextTacToe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
